@@ -12,6 +12,7 @@
  * Copyright  2006 Stefan Bethge
  * Copyright  2006 Christian Thaeter
  * Copyright  2007 Joseph P. Skudlarek
+ * Copyright  2008 Fedor P. Goncharov
  *
  * Permission to use, copy, modify, distribute, and sell this software
  * and its documentation for any purpose is hereby granted without
@@ -46,6 +47,7 @@
  *      Henry Davies <hdavies@ameritech.net> for the
  *      Linuxcare Inc. David Kennedy <dkennedy@linuxcare.com>
  *      Fred Hucht <fred@thp.Uni-Duisburg.de>
+ *      Fedor P. Goncharov <fedgo@gorodok.net>
  *
  * Trademarks are the property of their respective owners.
  */
@@ -69,9 +71,11 @@
 #include <xf86_OSproc.h>
 #include <xf86Xinput.h>
 #include "mipointer.h"
+#include <exevents.h>
 
-#define SYNAPTICS_PRIVATE
 #include "synaptics.h"
+#include "synapticsstr.h"
+#include "synaptics-properties.h"
 
 typedef enum {
     BOTTOM_EDGE = 1,
@@ -120,6 +124,10 @@ static Bool DeviceOff(DeviceIntPtr);
 static Bool DeviceClose(DeviceIntPtr);
 static Bool QueryHardware(LocalDevicePtr);
 
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 3
+void InitDeviceProperties(LocalDevicePtr local);
+Bool SetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop);
+#endif
 
 InputDriverRec SYNAPTICS = {
     1,
@@ -136,8 +144,8 @@ static XF86ModuleVersionInfo VersionRec = {
     MODULEVENDORSTRING,
     MODINFOSTRING1,
     MODINFOSTRING2,
-    XF86_VERSION_CURRENT,
-    1, 0, 0,
+    XORG_VERSION_CURRENT,
+    PACKAGE_VERSION_MAJOR, PACKAGE_VERSION_MINOR, PACKAGE_VERSION_PATCHLEVEL,
     ABI_CLASS_XINPUT,
     ABI_XINPUT_VERSION,
     MOD_CLASS_XINPUT,
@@ -168,11 +176,13 @@ SetDeviceAndProtocol(LocalDevicePtr local)
     device = xf86FindOptionValue(local->options, "Device");
     if (!device) {
 	device = xf86FindOptionValue(local->options, "Path");
-	xf86SetStrOption(local->options, "Device", device);
+	xf86ReplaceStrOption(local->options, "Device", device);
     }
     if (device && strstr(device, "/dev/input/event")) {
+#ifdef BUILD_EVENTCOMM
 	/* trust the device name if we've been given one */
 	proto = SYN_PROTO_EVENT;
+#endif
     } else {
 	str_par = xf86FindOptionValue(local->options, "Protocol");
 	if (str_par && !strcmp(str_par, "psaux")) {
@@ -188,8 +198,10 @@ SetDeviceAndProtocol(LocalDevicePtr local)
 	} else if (str_par && !strcmp(str_par, "alps")) {
 	    proto = SYN_PROTO_ALPS;
 	} else { /* default to auto-dev */
+#ifdef BUILD_EVENTCOMM
 	    if (event_proto_operations.AutoDevProbe(local))
 		proto = SYN_PROTO_EVENT;
+#endif
 	}
     }
     switch (proto) {
@@ -380,7 +392,7 @@ SynapticsPreInit(InputDriverPtr drv, IDevPtr dev, int flags)
     pars->finger_high = xf86SetIntOption(opts, "FingerHigh", 30);
     pars->finger_press = xf86SetIntOption(opts, "FingerPress", 256);
     pars->tap_time = xf86SetIntOption(opts, "MaxTapTime", 180);
-    pars->tap_move = xf86SetIntOption(opts, "MaxTapMove", 220);
+    pars->tap_move = xf86SetIntOption(opts, "MaxTapMove", 25);
     pars->tap_time_2 = xf86SetIntOption(opts, "MaxDoubleTapTime", 180);
     pars->click_time = xf86SetIntOption(opts, "ClickTime", 100);
     pars->fast_taps = xf86SetIntOption(opts, "FastTaps", FALSE);
@@ -389,6 +401,7 @@ SynapticsPreInit(InputDriverPtr drv, IDevPtr dev, int flags)
     pars->scroll_dist_vert = xf86SetIntOption(opts, "VertScrollDelta", 100);
     pars->scroll_dist_horiz = xf86SetIntOption(opts, "HorizScrollDelta", 100);
     pars->scroll_edge_vert = xf86SetBoolOption(opts, "VertEdgeScroll", TRUE);
+    pars->special_scroll_area_right  = xf86SetBoolOption(opts, "SpecialScrollAreaRight", TRUE);
     pars->scroll_edge_horiz = xf86SetBoolOption(opts, "HorizEdgeScroll", TRUE);
     pars->scroll_edge_corner = xf86SetBoolOption(opts, "CornerCoasting", FALSE);
     pars->scroll_twofinger_vert = xf86SetBoolOption(opts, "VertTwoFingerScroll", FALSE);
@@ -408,13 +421,13 @@ SynapticsPreInit(InputDriverPtr drv, IDevPtr dev, int flags)
     pars->guestmouse_off = xf86SetBoolOption(opts, "GuestMouseOff", FALSE);
     pars->locked_drags = xf86SetBoolOption(opts, "LockedDrags", FALSE);
     pars->locked_drag_time = xf86SetIntOption(opts, "LockedDragTimeout", 5000);
-    pars->tap_action[RT_TAP] = xf86SetIntOption(opts, "RTCornerButton", 0);
-    pars->tap_action[RB_TAP] = xf86SetIntOption(opts, "RBCornerButton", 0);
+    pars->tap_action[RT_TAP] = xf86SetIntOption(opts, "RTCornerButton", 2);
+    pars->tap_action[RB_TAP] = xf86SetIntOption(opts, "RBCornerButton", 3);
     pars->tap_action[LT_TAP] = xf86SetIntOption(opts, "LTCornerButton", 0);
     pars->tap_action[LB_TAP] = xf86SetIntOption(opts, "LBCornerButton", 0);
-    pars->tap_action[F1_TAP] = xf86SetIntOption(opts, "TapButton1",     0);
-    pars->tap_action[F2_TAP] = xf86SetIntOption(opts, "TapButton2",     0);
-    pars->tap_action[F3_TAP] = xf86SetIntOption(opts, "TapButton3",     0);
+    pars->tap_action[F1_TAP] = xf86SetIntOption(opts, "TapButton1",     1);
+    pars->tap_action[F2_TAP] = xf86SetIntOption(opts, "TapButton2",     2);
+    pars->tap_action[F3_TAP] = xf86SetIntOption(opts, "TapButton3",     3);
     pars->click_action[F1_CLICK1] = xf86SetIntOption(opts, "ClickFinger1",     1);
     pars->click_action[F2_CLICK1] = xf86SetIntOption(opts, "ClickFinger2",     1);
     pars->click_action[F3_CLICK1] = xf86SetIntOption(opts, "ClickFinger3",     1);
@@ -491,6 +504,7 @@ SynapticsPreInit(InputDriverPtr drv, IDevPtr dev, int flags)
 	xf86CloseSerial(local->fd);
     }
     local->fd = -1;
+
     return local;
 
  SetupProc_fail:
@@ -660,6 +674,11 @@ DeviceInit(DeviceIntPtr dev)
     if (!alloc_param_data(local))
 	return !Success;
 
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 3
+    InitDeviceProperties(local);
+    XIRegisterPropertyHandler(local->dev, SetProperty, NULL);
+#endif
+
     return Success;
 }
 
@@ -749,7 +768,7 @@ edge_detection(SynapticsPrivate *priv, int x, int y)
     if (priv->synpara->circular_pad)
 	return circular_edge_detection(priv, x, y);
 
-    if (x > priv->synpara->right_edge)
+    if (x >= priv->synpara->right_edge)
 	edge |= RIGHT_EDGE;
     else if (x < priv->synpara->left_edge)
 	edge |= LEFT_EDGE;
@@ -1109,8 +1128,7 @@ HandleTapProcessing(SynapticsPrivate *priv, struct SynapticsHwState *hw,
 
     touch = finger && !priv->finger_state;
     release = !finger && priv->finger_state;
-    move = ((priv->tap_max_fingers <= 1) &&
-	    ((abs(hw->x - priv->touch_on.x) >= para->tap_move) ||
+    move = (((abs(hw->x - priv->touch_on.x) >= para->tap_move) ||
 	     (abs(hw->y - priv->touch_on.y) >= para->tap_move)));
 
     if (touch) {
@@ -1702,7 +1720,7 @@ HandleScrolling(SynapticsPrivate *priv, struct SynapticsHwState *hw,
     return delay;
 }
 
-static int
+static void
 HandleClickWithFingers(SynapticsSHM *para, struct SynapticsHwState *hw)
 {
     int action = 0;
@@ -1828,14 +1846,32 @@ HandleState(LocalDevicePtr local, struct SynapticsHwState *hw)
 
     /*
      * Some touchpads have a scroll wheel region where a very large X
-     * coordinate is reported. For such touchpads, we adjust the X
-     * coordinate to eliminate the discontinuity.
+     * coordinate is reported. 
+     *
+     *    We suggest two  solution this problem:
      */
     if (hw->x <= XMAX_VALID) {
 	if (priv->largest_valid_x < hw->x)
 	    priv->largest_valid_x = hw->x;
     } else {
-	hw->x = priv->largest_valid_x;
+      if (!(para->special_scroll_area_right))
+      /*               First:
+      * Adjust the X coordinate to eliminate the discontinuity 
+      * and use it region as 1 coordinate size line.
+      */
+	hw->x = priv->largest_valid_x + 1;
+      else {
+      /*               Second (default):
+       * Adjust the X coordinate to eliminate the discontinuity
+       * and use it region as scroll area automaticly.
+       */	
+	
+	if (priv->synpara->right_edge > priv->largest_valid_x + 1)
+	  priv->synpara->right_edge=priv->largest_valid_x + 1;
+	para->special_scroll_area_right = FALSE;
+       
+	hw->x = priv->largest_valid_x + 1;
+      }
     }
 
     edge = edge_detection(priv, hw->x, hw->y);
@@ -2024,3 +2060,4 @@ QueryHardware(LocalDevicePtr local)
     priv->proto_ops->DeviceOffHook(local);
     return TRUE;
 }
+
